//This file is a part of xboa
//
//xboa is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//xboa is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with xboa in the doc folder.  If not, see 
//<http://www.gnu.org/licenses/>.
//


/*
Bunchcore.h
Set of core functions for Bunch python module
I want to improve speed - the thing I really worry about is the time taken 
to calc moments. My optimisation is to use Hitcore.get_x() etc directly
rather than requiring the python type lookups etc

Bunchcore is a list of Hit PyObjects and corresponding Hitcore objects
Nb I store separate arrays of Hitcore and Hit objects for speed reasons
but the Hitcore and Hit objects should always correspond!
*/

#ifndef Bunchcore_h 
#define Bunchcore_h

#include <Python.h>
#include "Hitcore.h"

static PyTypeObject      BunchcoreType;
static PySequenceMethods Bunchcore_as_seq;

//I want to be able to directly access the hitcore without going through Python gumph
typedef struct  { PyObject* hit; Hitcore* hitcore; } Bunchcore_hit; 
//Bunchcore data struct is a vector of Bunchcore_hits
typedef struct { PyObject_HEAD; vector* vec; } Bunchcore;

//deallocate memory from Bunchcore
static void Bunchcore_dealloc(Bunchcore * self);
//initialise Bunchcore objects
static int  Bunchcore_init(Bunchcore* self, int length);
//reallocate memory (extend)
static Bunchcore* bc_realloc(Bunchcore* self, int new_length);

static int length(Bunchcore* self) {return self->vec->data_length/self->vec->object_size;}
static PyObject*  length_python(PyObject* self, PyObject* args);
static Py_ssize_t len_python(PyObject* self); //len() function wants Py_ssize_t in return

////// GET METHODS ///////
static PyObject* get_item_python(PyObject* self, PyObject* args);

////// SET METHODS ////////
static void      set_item_bc_hit(Bunchcore* self, PyObject* a_hit, Hitcore* a_hitcore, int pos, int* ierr);
static PyObject* set_item_python(PyObject* self, PyObject* args);

/////// OTHER STUFF ///////
//get moments using char list to index moment type
static double    get_moment_char       (Bunchcore*  self, const char** axes, double* means, int n_axes, int* ierr);
static void covariance_matrix_char(Bunchcore*  self, const char* axes[], double* means, int n_axes, double* target, int* ierr);

//get moments using PyObjects
//helper function to parse list of strings - USER must FREE const char** afterwards
static void make_array_from_char_list (PyObject* py_list, const char*** string_list, int* list_size);
static void make_array_from_float_list(PyObject* py_list, double** double_list,      int* list_size);
//calculate the moment
static PyObject*    get_moment_python        (PyObject* self, PyObject* args);
static PyObject*    covariance_matrix_python (PyObject* self, PyObject* args);

// inner loop for cut (speed optimisation); returns 0 on success, 1 on error
static int _cut_double(Bunchcore* self, const char* variable,
                  PyObject* comp, const double cut_value,
                  const int is_local);
static PyObject* _cut_double_python(PyObject* self, PyObject* args);

void no_warn(void);

/////// API ////////

//initialise Bunchcore python module
PyMODINIT_FUNC   initBunchcore (void);

//hitcore api stuff
void**               hitcore_api;
void                 import_hitcore(void);

hc_get_dbl_function  (*hc_get_dbl_function_by_string)(const char*);
static void          (*hc_set_local_weight)(Hitcore* hc, double weight); 
static void          (*hc_set_global_weight)(Hitcore* hc, double weight); 

//vector (in Hitcore.h)
static vector* (*vector_alloc  )(size_t object_size, int n_elements);
static vector* (*vector_realloc)(vector* vec, int n_elements);
static vector* (*vector_init   )(void* data, size_t data_length, size_t allocd_length, size_t object_size);
static vector* (*vector_free   )(vector* vec);
static void    (*vector_insert )(vector* vec, void* start, void* end, size_t insert_point, int* ierr);
static void*   (*vector_el     )(vector* vec, int element);
static void*   (*vector_el_bc  )(vector* vec, int element);
static size_t  (*vector_size   )(vector* vec);

#endif
